@@ -43,6 +43,8 @@
 #include "common.h"
 #include "storage.h"
 #include "idmap.h"
+#include "simutil.h"
+#include "util.h"
 
 #define GPRS_FLAG_ATTACHING 0x1
 #define GPRS_FLAG_RECHECK 0x2
@@ -58,6 +60,18 @@
 static GSList *g_drivers = NULL;
 static GSList *g_context_drivers = NULL;
 
+/* 27.007 Section 7.29 */
+enum packet_bearer {
+	PACKET_BEARER_NONE =		0,
+	PACKET_BEARER_GPRS =		1,
+	PACKET_BEARER_EGPRS =		2,
+	PACKET_BEARER_UMTS =		3,
+	PACKET_BEARER_HSUPA =		4,
+	PACKET_BEARER_HSDPA =		5,
+	PACKET_BEARER_HSUPA_HSDPA =	6,
+	PACKET_BEARER_EPS =		7,
+};
+
 struct ofono_gprs {
 	GSList *contexts;
 	ofono_bool_t attached;
@@ -83,6 +97,29 @@ struct ofono_gprs {
 	const struct ofono_gprs_driver *driver;
 	void *driver_data;
 	struct ofono_atom *atom;
+	struct ofono_sim_context *sim_context;
+};
+
+struct ipv4_settings {
+	gboolean static_ip;
+	char *ip;
+	char *netmask;
+	char *gateway;
+	char **dns;
+	char *proxy;
+};
+
+struct ipv6_settings {
+	char *ip;
+	unsigned char prefix_len;
+	char *gateway;
+	char **dns;
+};
+
+struct context_settings {
+	char *interface;
+	struct ipv4_settings *ipv4;
+	struct ipv6_settings *ipv6;
 };
 
 struct ofono_gprs_context {
@@ -91,20 +128,10 @@ struct ofono_gprs_context {
 	ofono_bool_t inuse;
 	const struct ofono_gprs_context_driver *driver;
 	void *driver_data;
+	struct context_settings *settings;
 	struct ofono_atom *atom;
 };
 
-struct context_settings {
-	enum ofono_gprs_context_type type;
-	char *interface;
-	gboolean static_ip;
-	char *ip;
-	char *netmask;
-	char *gateway;
-	char **dns;
-	char *proxy;
-};
-
 struct pri_context {
 	ofono_bool_t active;
 	enum ofono_gprs_context_type type;
@@ -114,7 +141,6 @@ struct pri_context {
 	unsigned int id;
 	char *path;
 	char *key;
-	struct context_settings *settings;
 	char *proxy_host;
 	uint16_t proxy_port;
 	DBusMessage *pending;
@@ -126,6 +152,29 @@ struct pri_context {
 static void gprs_netreg_update(struct ofono_gprs *gprs);
 static void gprs_deactivate_next(struct ofono_gprs *gprs);
 
+const char *packet_bearer_to_string(int bearer)
+{
+	switch (bearer) {
+	case PACKET_BEARER_NONE:
+		return "none";
+	case PACKET_BEARER_GPRS:
+		return "gprs";
+	case PACKET_BEARER_EGPRS:
+		return "edge";
+	case PACKET_BEARER_UMTS:
+		return "umts";
+	case PACKET_BEARER_HSUPA:
+		return "hsupa";
+	case PACKET_BEARER_HSDPA:
+		return "hsdpa";
+	case PACKET_BEARER_HSUPA_HSDPA:
+		return "hspa";
+	case PACKET_BEARER_EPS:
+		return "lte";
+	}
+	return "";
+}
+
 static const char *gprs_context_default_name(enum ofono_gprs_context_type type)
 {
 	switch (type) {
@@ -144,7 +193,8 @@ static const char *gprs_context_default_name(enum ofono_gprs_context_type type)
 	return NULL;
 }
 
-static const char *gprs_context_type_to_string(enum ofono_gprs_context_type type)
+static const char *gprs_context_type_to_string(
+					enum ofono_gprs_context_type type)
 {
 	switch (type) {
 	case OFONO_GPRS_CONTEXT_TYPE_ANY:
@@ -189,6 +239,8 @@ static const char *gprs_proto_to_string(enum ofono_gprs_proto proto)
 		return "ip";
 	case OFONO_GPRS_PROTO_IPV6:
 		return "ipv6";
+	case OFONO_GPRS_PROTO_IPV4V6:
+		return "dual";
 	};
 
 	return NULL;
@@ -203,6 +255,9 @@ static gboolean gprs_proto_from_string(const char *str,
 	} else if (g_str_equal(str, "ipv6")) {
 		*proto = OFONO_GPRS_PROTO_IPV6;
 		return TRUE;
+	} else if (g_str_equal(str, "dual")) {
+		*proto = OFONO_GPRS_PROTO_IPV4V6;
+		return TRUE;
 	}
 
 	return FALSE;
@@ -218,6 +273,67 @@ static void gprs_cid_release(struct ofono_gprs *gprs, unsigned int id)
 	idmap_put(gprs->cid_map, id);
 }
 
+static gboolean assign_context(struct pri_context *ctx)
+{
+	struct idmap *cidmap = ctx->gprs->cid_map;
+	unsigned int cid_min;
+	GSList *l;
+
+	if (cidmap == NULL)
+		return FALSE;
+
+	cid_min = idmap_get_min(cidmap);
+
+	ctx->context.cid = gprs_cid_alloc(ctx->gprs);
+	if (ctx->context.cid == 0)
+		return FALSE;
+
+	for (l = ctx->gprs->context_drivers; l; l = l->next) {
+		struct ofono_gprs_context *gc = l->data;
+
+		if (gc->inuse == TRUE)
+			continue;
+
+		if (gc->driver == NULL)
+			continue;
+
+		if (gc->driver->activate_primary == NULL ||
+				gc->driver->deactivate_primary == NULL)
+			continue;
+
+		if (gc->type != OFONO_GPRS_CONTEXT_TYPE_ANY &&
+				gc->type != ctx->type)
+			continue;
+
+		ctx->context_driver = gc;
+		ctx->context_driver->inuse = TRUE;
+
+		if (ctx->context.proto == OFONO_GPRS_PROTO_IPV4V6 ||
+				ctx->context.proto == OFONO_GPRS_PROTO_IP)
+			gc->settings->ipv4 = g_new0(struct ipv4_settings, 1);
+
+		if (ctx->context.proto == OFONO_GPRS_PROTO_IPV4V6 ||
+				ctx->context.proto == OFONO_GPRS_PROTO_IPV6)
+			gc->settings->ipv6 = g_new0(struct ipv6_settings, 1);
+
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static void release_context(struct pri_context *ctx)
+{
+	if (ctx == NULL || ctx->gprs == NULL || ctx->context_driver == NULL)
+		return;
+
+	gprs_cid_release(ctx->gprs, ctx->context.cid);
+	ctx->context.cid = 0;
+	ctx->context_driver->inuse = FALSE;
+	ctx->context_driver = NULL;
+	ctx->active = FALSE;
+}
+
 static struct pri_context *gprs_context_by_path(struct ofono_gprs *gprs,
 						const char *ctx_path)
 {
@@ -235,17 +351,31 @@ static struct pri_context *gprs_context_by_path(struct ofono_gprs *gprs,
 
 static void context_settings_free(struct context_settings *settings)
 {
-	g_free(settings->interface);
-	g_free(settings->ip);
-	g_free(settings->netmask);
-	g_free(settings->gateway);
-	g_strfreev(settings->dns);
-	g_free(settings->proxy);
+	if (settings->ipv4) {
+		g_free(settings->ipv4->ip);
+		g_free(settings->ipv4->netmask);
+		g_free(settings->ipv4->gateway);
+		g_strfreev(settings->ipv4->dns);
+		g_free(settings->ipv4->proxy);
 
-	g_free(settings);
+		g_free(settings->ipv4);
+		settings->ipv4 = NULL;
+	}
+
+	if (settings->ipv6) {
+		g_free(settings->ipv6->ip);
+		g_free(settings->ipv6->gateway);
+		g_strfreev(settings->ipv6->dns);
+
+		g_free(settings->ipv6);
+		settings->ipv6 = NULL;
+	}
+
+	g_free(settings->interface);
+	settings->interface = NULL;
 }
 
-static void context_settings_append_variant(struct context_settings *settings,
+static void context_settings_append_ipv4(struct context_settings *settings,
 						DBusMessageIter *iter)
 {
 	DBusMessageIter variant;
@@ -266,42 +396,42 @@ static void context_settings_append_variant(struct context_settings *settings,
 
 	dbus_message_iter_open_container(&variant, DBUS_TYPE_ARRAY,
 						typesig, &array);
-	if (settings == NULL)
+	if (settings == NULL || settings->ipv4 == NULL)
 		goto done;
 
 	ofono_dbus_dict_append(&array, "Interface",
 				DBUS_TYPE_STRING, &settings->interface);
 
-	if (settings->type == OFONO_GPRS_CONTEXT_TYPE_MMS) {
-		if (settings->proxy)
-			ofono_dbus_dict_append(&array, "Proxy",
-					DBUS_TYPE_STRING, &settings->proxy);
+	/* If we have a Proxy, no other settings are relevant */
+	if (settings->ipv4->proxy) {
+		ofono_dbus_dict_append(&array, "Proxy", DBUS_TYPE_STRING,
+					&settings->ipv4->proxy);
 		goto done;
 	}
 
-	if (settings->static_ip == TRUE)
+	if (settings->ipv4->static_ip == TRUE)
 		method = "static";
 	else
 		method = "dhcp";
 
 	ofono_dbus_dict_append(&array, "Method", DBUS_TYPE_STRING, &method);
 
-	if (settings->ip)
+	if (settings->ipv4->ip)
 		ofono_dbus_dict_append(&array, "Address", DBUS_TYPE_STRING,
-					&settings->ip);
+					&settings->ipv4->ip);
 
-	if (settings->netmask)
+	if (settings->ipv4->netmask)
 		ofono_dbus_dict_append(&array, "Netmask", DBUS_TYPE_STRING,
-					&settings->netmask);
+					&settings->ipv4->netmask);
 
-	if (settings->gateway)
+	if (settings->ipv4->gateway)
 		ofono_dbus_dict_append(&array, "Gateway", DBUS_TYPE_STRING,
-					&settings->gateway);
+					&settings->ipv4->gateway);
 
-	if (settings->dns)
+	if (settings->ipv4->dns)
 		ofono_dbus_dict_append_array(&array, "DomainNameServers",
 						DBUS_TYPE_STRING,
-						&settings->dns);
+						&settings->ipv4->dns);
 
 done:
 	dbus_message_iter_close_container(&variant, &array);
@@ -309,7 +439,7 @@ done:
 	dbus_message_iter_close_container(iter, &variant);
 }
 
-static void context_settings_append_dict(struct context_settings *settings,
+static void context_settings_append_ipv4_dict(struct context_settings *settings,
 						DBusMessageIter *dict)
 {
 	DBusMessageIter entry;
@@ -320,18 +450,85 @@ static void context_settings_append_dict(struct context_settings *settings,
 
 	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
 
-	context_settings_append_variant(settings, &entry);
+	context_settings_append_ipv4(settings, &entry);
 
 	dbus_message_iter_close_container(dict, &entry);
 }
 
-static void pri_context_signal_settings(struct pri_context *ctx)
+static void context_settings_append_ipv6(struct context_settings *settings,
+						DBusMessageIter *iter)
+{
+	DBusMessageIter variant;
+	DBusMessageIter array;
+	char typesig[5];
+	char arraysig[6];
+
+	arraysig[0] = DBUS_TYPE_ARRAY;
+	arraysig[1] = typesig[0] = DBUS_DICT_ENTRY_BEGIN_CHAR;
+	arraysig[2] = typesig[1] = DBUS_TYPE_STRING;
+	arraysig[3] = typesig[2] = DBUS_TYPE_VARIANT;
+	arraysig[4] = typesig[3] = DBUS_DICT_ENTRY_END_CHAR;
+	arraysig[5] = typesig[4] = '\0';
+
+	dbus_message_iter_open_container(iter, DBUS_TYPE_VARIANT,
+						arraysig, &variant);
+
+	dbus_message_iter_open_container(&variant, DBUS_TYPE_ARRAY,
+						typesig, &array);
+	if (settings == NULL || settings->ipv6 == NULL)
+		goto done;
+
+	ofono_dbus_dict_append(&array, "Interface",
+				DBUS_TYPE_STRING, &settings->interface);
+
+	if (settings->ipv6->ip)
+		ofono_dbus_dict_append(&array, "Address", DBUS_TYPE_STRING,
+					&settings->ipv6->ip);
+
+	if (settings->ipv6->prefix_len)
+		ofono_dbus_dict_append(&array, "PrefixLength", DBUS_TYPE_BYTE,
+					&settings->ipv6->prefix_len);
+
+	if (settings->ipv6->gateway)
+		ofono_dbus_dict_append(&array, "Gateway", DBUS_TYPE_STRING,
+					&settings->ipv6->gateway);
+
+	if (settings->ipv6->dns)
+		ofono_dbus_dict_append_array(&array, "DomainNameServers",
+						DBUS_TYPE_STRING,
+						&settings->ipv6->dns);
+
+done:
+	dbus_message_iter_close_container(&variant, &array);
+
+	dbus_message_iter_close_container(iter, &variant);
+}
+
+static void context_settings_append_ipv6_dict(struct context_settings *settings,
+						DBusMessageIter *dict)
+{
+	DBusMessageIter entry;
+	const char *key = "IPv6.Settings";
+
+	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
+						NULL, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	context_settings_append_ipv6(settings, &entry);
+
+	dbus_message_iter_close_container(dict, &entry);
+}
+
+static void signal_settings(struct pri_context *ctx, const char *prop,
+		void (*append)(struct context_settings *, DBusMessageIter *))
+
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path = ctx->path;
 	DBusMessage *signal;
 	DBusMessageIter iter;
-	const char *prop = "Settings";
+	struct context_settings *settings;
 
 	signal = dbus_message_new_signal(path,
 					OFONO_CONNECTION_CONTEXT_INTERFACE,
@@ -341,14 +538,29 @@ static void pri_context_signal_settings(struct pri_context *ctx)
 		return;
 
 	dbus_message_iter_init_append(signal, &iter);
-
 	dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &prop);
 
-	context_settings_append_variant(ctx->settings, &iter);
+	if (ctx->context_driver)
+		settings = ctx->context_driver->settings;
+	else
+		settings = NULL;
 
+	append(settings, &iter);
 	g_dbus_send_message(conn, signal);
 }
 
+static void pri_context_signal_settings(struct pri_context *ctx,
+					gboolean ipv4, gboolean ipv6)
+{
+	if (ipv4)
+		signal_settings(ctx, "Settings",
+				context_settings_append_ipv4);
+
+	if (ipv6)
+		signal_settings(ctx, "IPv6.Settings",
+				context_settings_append_ipv6);
+}
+
 static void pri_parse_proxy(struct pri_context *ctx, const char *proxy)
 {
 	char *scheme, *host, *port, *path;
@@ -431,7 +643,7 @@ done:
 	close(sk);
 }
 
-static void pri_setaddr(const char *interface, const char *address)
+static void pri_set_ipv4_addr(const char *interface, const char *address)
 {
 	struct ifreq ifr;
 	struct sockaddr_in addr;
@@ -515,21 +727,28 @@ static void pri_setproxy(const char *interface, const char *proxy)
 
 static void pri_reset_context_settings(struct pri_context *ctx)
 {
+	struct context_settings *settings;
 	char *interface;
+	gboolean signal_ipv4;
+	gboolean signal_ipv6;
 
-	if (ctx->settings == NULL)
+	if (ctx->context_driver == NULL)
 		return;
 
-	interface = ctx->settings->interface;
-	ctx->settings->interface = NULL;
+	settings = ctx->context_driver->settings;
 
-	context_settings_free(ctx->settings);
-	ctx->settings = NULL;
+	interface = settings->interface;
+	settings->interface = NULL;
 
-	pri_context_signal_settings(ctx);
+	signal_ipv4 = settings->ipv4 != NULL;
+	signal_ipv6 = settings->ipv6 != NULL;
+
+	context_settings_free(settings);
+
+	pri_context_signal_settings(ctx, signal_ipv4, signal_ipv6);
 
 	if (ctx->type == OFONO_GPRS_CONTEXT_TYPE_MMS) {
-		pri_setaddr(interface, NULL);
+		pri_set_ipv4_addr(interface, NULL);
 
 		g_free(ctx->proxy_host);
 		ctx->proxy_host = NULL;
@@ -541,45 +760,22 @@ static void pri_reset_context_settings(struct pri_context *ctx)
 	g_free(interface);
 }
 
-static void pri_update_context_settings(struct pri_context *ctx,
-					const char *interface,
-					ofono_bool_t static_ip,
-					const char *ip, const char *netmask,
-					const char *gateway, const char **dns)
+static void pri_update_mms_context_settings(struct pri_context *ctx)
 {
-	if (ctx->settings)
-		context_settings_free(ctx->settings);
+	struct ofono_gprs_context *gc = ctx->context_driver;
+	struct context_settings *settings = gc->settings;
 
-	ctx->settings = g_try_new0(struct context_settings, 1);
-	if (ctx->settings == NULL)
-		return;
+	if (ctx->message_proxy)
+		settings->ipv4->proxy = g_strdup(ctx->message_proxy);
 
-	ctx->settings->type = ctx->type;
+	pri_parse_proxy(ctx, ctx->message_proxy);
 
-	ctx->settings->interface = g_strdup(interface);
-	ctx->settings->static_ip = static_ip;
-	ctx->settings->ip = g_strdup(ip);
-	ctx->settings->netmask = g_strdup(netmask);
-	ctx->settings->gateway = g_strdup(gateway);
-	ctx->settings->dns = g_strdupv((char **)dns);
+	DBG("proxy %s port %u", ctx->proxy_host, ctx->proxy_port);
 
-	if (ctx->type == OFONO_GPRS_CONTEXT_TYPE_MMS && ctx->message_proxy)
-		ctx->settings->proxy = g_strdup(ctx->message_proxy);
+	pri_set_ipv4_addr(settings->interface, settings->ipv4->ip);
 
-	pri_ifupdown(interface, TRUE);
-
-	if (ctx->type == OFONO_GPRS_CONTEXT_TYPE_MMS) {
-		pri_parse_proxy(ctx, ctx->message_proxy);
-
-		DBG("proxy %s port %u", ctx->proxy_host, ctx->proxy_port);
-
-		pri_setaddr(interface, ip);
-
-		if (ctx->proxy_host)
-			pri_setproxy(interface, ctx->proxy_host);
-	}
-
-	pri_context_signal_settings(ctx);
+	if (ctx->proxy_host)
+		pri_setproxy(settings->interface, ctx->proxy_host);
 }
 
 static void append_context_properties(struct pri_context *ctx,
@@ -590,6 +786,7 @@ static void append_context_properties(struct pri_context *ctx,
 	const char *name = ctx->name;
 	dbus_bool_t value;
 	const char *strvalue;
+	struct context_settings *settings;
 
 	ofono_dbus_dict_append(dict, "Name", DBUS_TYPE_STRING, &name);
 
@@ -622,7 +819,13 @@ static void append_context_properties(struct pri_context *ctx,
 					DBUS_TYPE_STRING, &strvalue);
 	}
 
-	context_settings_append_dict(ctx->settings, dict);
+	if (ctx->context_driver)
+		settings = ctx->context_driver->settings;
+	else
+		settings = NULL;
+
+	context_settings_append_ipv4_dict(settings, dict);
+	context_settings_append_ipv6_dict(settings, dict);
 }
 
 static DBusMessage *pri_get_properties(DBusConnection *conn,
@@ -648,30 +851,22 @@ static DBusMessage *pri_get_properties(DBusConnection *conn,
 	return reply;
 }
 
-static void pri_activate_callback(const struct ofono_error *error,
-					const char *interface,
-					ofono_bool_t static_ip,
-					const char *ip, const char *netmask,
-					const char *gateway, const char **dns,
-					void *data)
+static void pri_activate_callback(const struct ofono_error *error, void *data)
 {
 	struct pri_context *ctx = data;
+	struct ofono_gprs_context *gc = ctx->context_driver;
 	DBusConnection *conn = ofono_dbus_get_connection();
 	dbus_bool_t value;
 
-	DBG("%p %s", ctx, interface);
+	DBG("%p", ctx);
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		DBG("Activating context failed with error: %s",
 				telephony_error_to_str(error));
 		__ofono_dbus_pending_reply(&ctx->pending,
 					__ofono_error_failed(ctx->pending));
-
-		gprs_cid_release(ctx->gprs, ctx->context.cid);
-		ctx->context.cid = 0;
-		ctx->context_driver->inuse = FALSE;
-		ctx->context_driver = NULL;
-
+		context_settings_free(ctx->context_driver->settings);
+		release_context(ctx);
 		return;
 	}
 
@@ -679,13 +874,16 @@ static void pri_activate_callback(const struct ofono_error *error,
 	__ofono_dbus_pending_reply(&ctx->pending,
 				dbus_message_new_method_return(ctx->pending));
 
-	/*
-	 * If we don't have the interface, don't bother emitting any settings,
-	 * as nobody can make use of them
-	 */
-	if (interface != NULL)
-		pri_update_context_settings(ctx, interface, static_ip,
-						ip, netmask, gateway, dns);
+	if (gc->settings->interface != NULL) {
+		pri_ifupdown(gc->settings->interface, TRUE);
+
+		if (ctx->type == OFONO_GPRS_CONTEXT_TYPE_MMS &&
+				gc->settings->ipv4)
+			pri_update_mms_context_settings(ctx);
+
+		pri_context_signal_settings(ctx, gc->settings->ipv4 != NULL,
+						gc->settings->ipv6 != NULL);
+	}
 
 	value = ctx->active;
 	ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -707,16 +905,11 @@ static void pri_deactivate_callback(const struct ofono_error *error, void *data)
 		return;
 	}
 
-	gprs_cid_release(ctx->gprs, ctx->context.cid);
-	ctx->context.cid = 0;
-	ctx->active = FALSE;
-	ctx->context_driver->inuse = FALSE;
-	ctx->context_driver = NULL;
-
 	__ofono_dbus_pending_reply(&ctx->pending,
 				dbus_message_new_method_return(ctx->pending));
 
 	pri_reset_context_settings(ctx);
+	release_context(ctx);
 
 	value = ctx->active;
 	ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -958,38 +1151,6 @@ static DBusMessage *pri_set_message_center(struct pri_context *ctx,
 	return NULL;
 }
 
-static gboolean assign_context(struct pri_context *ctx)
-{
-	struct idmap *cidmap = ctx->gprs->cid_map;
-	unsigned int cid_min;
-	GSList *l;
-
-	if (cidmap == NULL)
-		return FALSE;
-
-	cid_min = idmap_get_min(cidmap);
-
-	ctx->context.cid = gprs_cid_alloc(ctx->gprs);
-	if (ctx->context.cid == 0)
-		return FALSE;
-
-	for (l = ctx->gprs->context_drivers; l; l = l->next) {
-		struct ofono_gprs_context *gc = l->data;
-
-		if (gc->inuse == TRUE)
-			continue;
-
-		if (gc->type == OFONO_GPRS_CONTEXT_TYPE_ANY ||
-						gc->type == ctx->type) {
-			ctx->context_driver = gc;
-			ctx->context_driver->inuse = TRUE;
-			return TRUE;
-		}
-	}
-
-	return FALSE;
-}
-
 static DBusMessage *pri_set_property(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -1037,16 +1198,10 @@ static DBusMessage *pri_set_property(DBusConnection *conn,
 		if (ctx->gprs->flags & GPRS_FLAG_ATTACHING)
 			return __ofono_error_attach_in_progress(msg);
 
-		if (value) {
-			if (assign_context(ctx) == FALSE)
-				return __ofono_error_not_implemented(msg);
-		}
+		if (value && assign_context(ctx) == FALSE)
+			return __ofono_error_not_implemented(msg);
 
 		gc = ctx->context_driver;
-		if (gc == NULL || gc->driver == NULL ||
-				gc->driver->activate_primary == NULL ||
-				gc->driver->deactivate_primary == NULL)
-			return __ofono_error_not_implemented(msg);
 
 		ctx->pending = dbus_message_ref(msg);
 
@@ -1170,15 +1325,8 @@ static void pri_context_destroy(gpointer userdata)
 {
 	struct pri_context *ctx = userdata;
 
-	if (ctx->settings) {
-		context_settings_free(ctx->settings);
-		ctx->settings = NULL;
-	}
-
 	g_free(ctx->proxy_host);
-
 	g_free(ctx->path);
-
 	g_free(ctx);
 }
 
@@ -1307,13 +1455,8 @@ static void gprs_attached_update(struct ofono_gprs *gprs)
 			if (ctx->active == FALSE)
 				continue;
 
-			gprs_cid_release(gprs, ctx->context.cid);
-			ctx->context.cid = 0;
-			ctx->active = FALSE;
-			ctx->context_driver->inuse = FALSE;
-			ctx->context_driver = NULL;
-
 			pri_reset_context_settings(ctx);
+			release_context(ctx);
 
 			value = FALSE;
 			ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -1373,6 +1516,18 @@ static void gprs_attach_callback(const struct ofono_error *error, void *data)
 	}
 }
 
+static void gprs_netreg_removed(struct ofono_gprs *gprs)
+{
+	gprs->netreg = NULL;
+
+	gprs->flags &= ~(GPRS_FLAG_RECHECK | GPRS_FLAG_ATTACHING);
+	gprs->status_watch = 0;
+	gprs->netreg_status = NETWORK_REGISTRATION_STATUS_NOT_REGISTERED;
+	gprs->driver_attached = FALSE;
+
+	gprs_attached_update(gprs);
+}
+
 static void gprs_netreg_update(struct ofono_gprs *gprs)
 {
 	ofono_bool_t attach;
@@ -1557,6 +1712,15 @@ static void write_context_settings(struct ofono_gprs *gprs,
 				gprs_context_type_to_string(context->type));
 	g_key_file_set_string(gprs->settings, context->key, "Protocol",
 				gprs_proto_to_string(context->context.proto));
+
+	if (context->type == OFONO_GPRS_CONTEXT_TYPE_MMS) {
+		g_key_file_set_string(gprs->settings, context->key,
+					"MessageProxy",
+					context->message_proxy);
+		g_key_file_set_string(gprs->settings, context->key,
+					"MessageCenter",
+					context->message_center);
+	}
 }
 
 static struct pri_context *add_context(struct ofono_gprs *gprs,
@@ -1576,6 +1740,7 @@ static struct pri_context *add_context(struct ofono_gprs *gprs,
 
 	context = pri_context_create(gprs, name, type);
 	if (context == NULL) {
+		idmap_put(gprs->pid_map, id);
 		ofono_error("Unable to allocate context struct");
 		return NULL;
 	}
@@ -1677,10 +1842,8 @@ static void gprs_deactivate_for_remove(const struct ofono_error *error,
 		return;
 	}
 
-	gprs_cid_release(gprs, ctx->context.cid);
-	ctx->context.cid = 0;
-	ctx->context_driver->inuse = FALSE;
-	ctx->context_driver = NULL;
+	pri_reset_context_settings(ctx);
+	release_context(ctx);
 
 	if (gprs->settings) {
 		g_key_file_remove_group(gprs->settings, ctx->key, NULL);
@@ -1772,13 +1935,8 @@ static void gprs_deactivate_for_all(const struct ofono_error *error,
 		return;
 	}
 
-	gprs_cid_release(gprs, ctx->context.cid);
-	ctx->active = FALSE;
-	ctx->context.cid = 0;
-	ctx->context_driver->inuse = FALSE;
-	ctx->context_driver = NULL;
-
 	pri_reset_context_settings(ctx);
+	release_context(ctx);
 
 	value = ctx->active;
 	conn = ofono_dbus_get_connection();
@@ -1974,6 +2132,12 @@ static void gprs_context_unregister(struct ofono_atom *atom)
 	if (gc->gprs == NULL)
 		return;
 
+	if (gc->settings) {
+		context_settings_free(gc->settings);
+		g_free(gc->settings);
+		gc->settings = NULL;
+	}
+
 	gc->gprs->context_drivers = g_slist_remove(gc->gprs->context_drivers,
 							gc);
 	gc->gprs = NULL;
@@ -1986,6 +2150,7 @@ void ofono_gprs_add_context(struct ofono_gprs *gprs,
 		return;
 
 	gc->gprs = gprs;
+	gc->settings = g_new0(struct context_settings, 1);
 
 	gprs->context_drivers = g_slist_append(gprs->context_drivers, gc);
 	__ofono_atom_register(gc->atom, gprs_context_unregister);
@@ -2028,13 +2193,8 @@ void ofono_gprs_context_deactivated(struct ofono_gprs_context *gc,
 		if (ctx->active == FALSE)
 			break;
 
-		gprs_cid_release(ctx->gprs, ctx->context.cid);
-		ctx->context.cid = 0;
-		ctx->active = FALSE;
-		ctx->context_driver->inuse = FALSE;
-		ctx->context_driver = NULL;
-
 		pri_reset_context_settings(ctx);
+		release_context(ctx);
 
 		value = FALSE;
 		ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -2143,6 +2303,112 @@ void ofono_gprs_context_set_type(struct ofono_gprs_context *gc,
 	gc->type = type;
 }
 
+void ofono_gprs_context_set_interface(struct ofono_gprs_context *gc,
+					const char *interface)
+{
+	struct context_settings *settings = gc->settings;
+
+	g_free(settings->interface);
+	settings->interface = g_strdup(interface);
+}
+
+void ofono_gprs_context_set_ipv4_address(struct ofono_gprs_context *gc,
+						const char *address,
+						gboolean static_ip)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv4 == NULL)
+		return;
+
+	g_free(settings->ipv4->ip);
+	settings->ipv4->ip = g_strdup(address);
+	settings->ipv4->static_ip = static_ip;
+}
+
+void ofono_gprs_context_set_ipv4_netmask(struct ofono_gprs_context *gc,
+						const char *netmask)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv4 == NULL)
+		return;
+
+	g_free(settings->ipv4->netmask);
+	settings->ipv4->netmask = g_strdup(netmask);
+}
+
+void ofono_gprs_context_set_ipv4_gateway(struct ofono_gprs_context *gc,
+						const char *gateway)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv4 == NULL)
+		return;
+
+	g_free(settings->ipv4->gateway);
+	settings->ipv4->gateway = g_strdup(gateway);
+}
+
+void ofono_gprs_context_set_ipv4_dns_servers(struct ofono_gprs_context *gc,
+						const char **dns)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv4 == NULL)
+		return;
+
+	g_strfreev(settings->ipv4->dns);
+	settings->ipv4->dns = g_strdupv((char **) dns);
+}
+
+void ofono_gprs_context_set_ipv6_address(struct ofono_gprs_context *gc,
+						const char *address)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv6 == NULL)
+		return;
+
+	g_free(settings->ipv6->ip);
+	settings->ipv6->ip = g_strdup(address);
+}
+
+void ofono_gprs_context_set_ipv6_prefix_length(struct ofono_gprs_context *gc,
+						unsigned char length)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv6 == NULL)
+		return;
+
+	settings->ipv6->prefix_len = length;
+}
+
+void ofono_gprs_context_set_ipv6_gateway(struct ofono_gprs_context *gc,
+						const char *gateway)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv6 == NULL)
+		return;
+
+	g_free(settings->ipv6->gateway);
+	settings->ipv6->gateway = g_strdup(gateway);
+}
+
+void ofono_gprs_context_set_ipv6_dns_servers(struct ofono_gprs_context *gc,
+						const char **dns)
+{
+	struct context_settings *settings = gc->settings;
+
+	if (settings->ipv6 == NULL)
+		return;
+
+	g_strfreev(settings->ipv6->dns);
+	settings->ipv6->dns = g_strdupv((char **) dns);
+}
+
 int ofono_gprs_driver_register(const struct ofono_gprs_driver *d)
 {
 	DBG("driver: %p, name: %s", d, d->name);
@@ -2162,12 +2428,8 @@ void ofono_gprs_driver_unregister(const struct ofono_gprs_driver *d)
 	g_drivers = g_slist_remove(g_drivers, (void *)d);
 }
 
-static void gprs_unregister(struct ofono_atom *atom)
+static void free_contexts(struct ofono_gprs *gprs)
 {
-	DBusConnection *conn = ofono_dbus_get_connection();
-	struct ofono_gprs *gprs = __ofono_atom_get_data(atom);
-	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
-	const char *path = __ofono_atom_get_path(atom);
 	GSList *l;
 
 	if (gprs->settings) {
@@ -2186,6 +2448,16 @@ static void gprs_unregister(struct ofono_atom *atom)
 	}
 
 	g_slist_free(gprs->contexts);
+}
+
+static void gprs_unregister(struct ofono_atom *atom)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_gprs *gprs = __ofono_atom_get_data(atom);
+	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
+	const char *path = __ofono_atom_get_path(atom);
+
+	free_contexts(gprs);
 
 	if (gprs->cid_map) {
 		idmap_free(gprs->cid_map);
@@ -2232,6 +2504,9 @@ static void gprs_remove(struct ofono_atom *atom)
 	if (gprs->driver && gprs->driver->remove)
 		gprs->driver->remove(gprs);
 
+	if (gprs->sim_context)
+		ofono_sim_context_free(gprs->sim_context);
+
 	g_free(gprs);
 }
 
@@ -2279,8 +2554,7 @@ static void netreg_watch(struct ofono_atom *atom,
 	struct ofono_gprs *gprs = data;
 
 	if (cond == OFONO_ATOM_WATCH_CONDITION_UNREGISTERED) {
-		gprs->status_watch = 0;
-		gprs->netreg = NULL;
+		gprs_netreg_removed(gprs);
 		return;
 	}
 
@@ -2422,7 +2696,7 @@ error:
 
 static void gprs_load_settings(struct ofono_gprs *gprs, const char *imsi)
 {
-	GError *error = NULL;
+	GError *error;
 	gboolean legacy = FALSE;
 	char **groups;
 	int i;
@@ -2434,6 +2708,7 @@ static void gprs_load_settings(struct ofono_gprs *gprs, const char *imsi)
 
 	gprs->imsi = g_strdup(imsi);
 
+	error = NULL;
 	gprs->powered = g_key_file_get_boolean(gprs->settings, SETTINGS_GROUP,
 						"Powered", &error);
 
@@ -2443,16 +2718,20 @@ static void gprs_load_settings(struct ofono_gprs *gprs, const char *imsi)
 	 * and RoamingAllowed = False
 	 */
 	if (error) {
+		g_error_free(error);
 		gprs->powered = TRUE;
 		g_key_file_set_boolean(gprs->settings, SETTINGS_GROUP,
 					"Powered", gprs->powered);
 	}
 
+	error = NULL;
 	gprs->roaming_allowed = g_key_file_get_boolean(gprs->settings,
 							SETTINGS_GROUP,
-							"RoamingAllowed", NULL);
+							"RoamingAllowed",
+							&error);
 
 	if (error) {
+		g_error_free(error);
 		gprs->roaming_allowed = FALSE;
 		g_key_file_set_boolean(gprs->settings, SETTINGS_GROUP,
 					"RoamingAllowed",
@@ -2485,13 +2764,114 @@ remove:
 		storage_sync(imsi, SETTINGS_STORE, gprs->settings);
 }
 
-void ofono_gprs_register(struct ofono_gprs *gprs)
+static void provision_context(const struct ofono_gprs_provision_data *ap,
+				struct ofono_gprs *gprs)
+{
+	unsigned int id;
+	struct pri_context *context = NULL;
+
+	/* Sanity check */
+	if (ap == NULL)
+		return;
+
+	if (ap->name == NULL || strlen(ap->name) > MAX_CONTEXT_NAME_LENGTH)
+		return;
+
+	if (ap->apn == NULL || strlen(ap->apn) > OFONO_GPRS_MAX_APN_LENGTH)
+		return;
+
+	if (is_valid_apn(ap->apn) == FALSE)
+		return;
+
+	if (ap->username &&
+			strlen(ap->username) > OFONO_GPRS_MAX_USERNAME_LENGTH)
+		return;
+
+	if (ap->password &&
+			strlen(ap->password) > OFONO_GPRS_MAX_PASSWORD_LENGTH)
+		return;
+
+	if (ap->message_proxy &&
+			strlen(ap->message_proxy) > MAX_MESSAGE_PROXY_LENGTH)
+		return;
+
+	if (ap->message_center &&
+			strlen(ap->message_center) > MAX_MESSAGE_CENTER_LENGTH)
+		return;
+
+	if (gprs->last_context_id)
+		id = idmap_alloc_next(gprs->pid_map, gprs->last_context_id);
+	else
+		id = idmap_alloc(gprs->pid_map);
+
+	if (id > idmap_get_max(gprs->pid_map))
+		return;
+
+	context = pri_context_create(gprs, ap->name, ap->type);
+	if (context == NULL) {
+		idmap_put(gprs->pid_map, id);
+		return;
+	}
+
+	context->id = id;
+
+	if (ap->username != NULL)
+		strcpy(context->context.username, ap->username);
+
+	if (ap->password != NULL)
+		strcpy(context->context.password, ap->password);
+
+	strcpy(context->context.apn, ap->apn);
+	context->context.proto = ap->proto;
+
+	if (ap->type == OFONO_GPRS_CONTEXT_TYPE_MMS) {
+		if (ap->message_proxy != NULL)
+			strcpy(context->message_proxy, ap->message_proxy);
+
+		if (ap->message_center != NULL)
+			strcpy(context->message_center, ap->message_center);
+	}
+
+	if (context_dbus_register(context) == FALSE)
+		return;
+
+	gprs->last_context_id = id;
+
+	if (gprs->settings) {
+		write_context_settings(gprs, context);
+		storage_sync(gprs->imsi, SETTINGS_STORE, gprs->settings);
+	}
+
+	gprs->contexts = g_slist_append(gprs->contexts, context);
+}
+
+static void provision_contexts(struct ofono_gprs *gprs, const char *mcc,
+				const char *mnc, const char *spn)
+{
+	struct ofono_gprs_provision_data *settings;
+	int count;
+	int i;
+
+	if (__ofono_gprs_provision_get_settings(mcc, mnc, spn,
+						&settings, &count) == FALSE) {
+		ofono_warn("Provisioning failed");
+		return;
+	}
+
+	for (i = 0; i < count; i++)
+		provision_context(&settings[i], gprs);
+
+	__ofono_gprs_provision_free_settings(settings, count);
+}
+
+static void ofono_gprs_finish_register(struct ofono_gprs *gprs)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	struct ofono_modem *modem = __ofono_atom_get_modem(gprs->atom);
 	const char *path = __ofono_atom_get_path(gprs->atom);
-	struct ofono_atom *netreg_atom;
-	struct ofono_atom *sim_atom;
+
+	if (gprs->contexts == NULL) /* Automatic provisioning failed */
+		add_context(gprs, NULL, OFONO_GPRS_CONTEXT_TYPE_INTERNET);
 
 	if (!g_dbus_register_interface(conn, path,
 					OFONO_CONNECTION_MANAGER_INTERFACE,
@@ -2500,37 +2880,73 @@ void ofono_gprs_register(struct ofono_gprs *gprs)
 		ofono_error("Could not create %s interface",
 				OFONO_CONNECTION_MANAGER_INTERFACE);
 
+		free_contexts(gprs);
 		return;
 	}
 
 	ofono_modem_add_interface(modem,
 				OFONO_CONNECTION_MANAGER_INTERFACE);
 
-	sim_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SIM);
-
-	if (sim_atom) {
-		struct ofono_sim *sim = __ofono_atom_get_data(sim_atom);
-		const char *imsi = ofono_sim_get_imsi(sim);
-
-		gprs_load_settings(gprs, imsi);
-	}
-
-	if (gprs->contexts == NULL)
-		add_context(gprs, NULL, OFONO_GPRS_CONTEXT_TYPE_INTERNET);
-
 	gprs->netreg_watch = __ofono_modem_add_atom_watch(modem,
 					OFONO_ATOM_TYPE_NETREG,
 					netreg_watch, gprs, NULL);
 
-	netreg_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_NETREG);
-
-	if (netreg_atom && __ofono_atom_get_registered(netreg_atom))
-		netreg_watch(netreg_atom,
-				OFONO_ATOM_WATCH_CONDITION_REGISTERED, gprs);
-
 	__ofono_atom_register(gprs->atom, gprs_unregister);
 }
 
+static void sim_spn_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_gprs *gprs	= userdata;
+	char *spn = NULL;
+	struct ofono_atom *sim_atom;
+	struct ofono_sim *sim = NULL;
+
+	if (ok)
+		spn = sim_string_to_utf8(data + 1, length - 1);
+
+	sim_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(gprs->atom),
+						OFONO_ATOM_TYPE_SIM);
+	if (sim_atom) {
+		sim = __ofono_atom_get_data(sim_atom);
+		provision_contexts(gprs, ofono_sim_get_mcc(sim),
+					ofono_sim_get_mnc(sim), spn);
+	}
+
+	g_free(spn);
+	ofono_gprs_finish_register(gprs);
+}
+
+void ofono_gprs_register(struct ofono_gprs *gprs)
+{
+	struct ofono_modem *modem = __ofono_atom_get_modem(gprs->atom);
+	struct ofono_atom *sim_atom;
+	struct ofono_sim *sim = NULL;
+
+	sim_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SIM);
+
+	if (sim_atom) {
+		const char *imsi;
+		sim = __ofono_atom_get_data(sim_atom);
+
+		imsi = ofono_sim_get_imsi(sim);
+		gprs_load_settings(gprs, imsi);
+	}
+
+	if (gprs->contexts == NULL && sim != NULL) {
+		/* Get Service Provider Name from SIM for provisioning */
+		gprs->sim_context = ofono_sim_context_create(sim);
+
+		if (ofono_sim_read(gprs->sim_context, SIM_EFSPN_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+					sim_spn_read_cb, gprs) >= 0)
+			return;
+	}
+
+	ofono_gprs_finish_register(gprs);
+}
+
 void ofono_gprs_remove(struct ofono_gprs *gprs)
 {
 	__ofono_atom_free(gprs->atom);
